#include<stdio.h>
#include<sys/types.h>
#include<sys/socket.h>
#include<sys/select.h>
#include<sys/time.h>
#include<stdlib.h>
#include<unistd.h>
#include<arpa/inet.h>
#include<netinet/in.h>
#include<string.h>
#include<netdb.h>
#include <fcntl.h>
#include <stdarg.h>

#define ERR_SOCKET_WRITE_ERROR 1000
#define ERR_SOCKET_READ_ERROR 1001
#define ERR_SOCKET_TIMEOUT 1002
#define ERR_SOCKET_SELECT 1003
#define BUFFER_SIZE_LINE 150
#define BUFFER_SIZE 8000
#define FTP_DEFAULT_PORT 21
#define ERR_CANT_CONNECT 2000
#define ERR_FTP_SERVER_CANT_CONNECT 3000

/*
Les commandes FTP sont spcifies ci-dessous :



      USER <SP> <nom d'utilisateur> <CRLF>
      PASS <SP> <mot de passe> <CRLF>
      ACCT <SP> <compte utilisateur> <CRLF>
      CWD <SP> <chemin d'accs> <CRLF>
      CDUP <CRLF>
      SMNT <SP> <chemin d'accs> <CRLF>
      QUIT <CRLF>
      REIN <CRLF>
      PORT <SP> <port> <CRLF>
      PASV <CRLF>
      TYPE <SP> <code type> <CRLF>
      STRU <SP> <code structure> <CRLF>
      MODE <SP> <code mode> <CRLF>
      RETR <SP> <chemin d'accs> <CRLF>
      STOR <SP> <chemin d'accs> <CRLF>
      STOU <CRLF>
      APPE <SP> <chemin d'accs> <CRLF>
      ALLO <SP> <entier dcimal>
        [<SP> R <SP> <entier dcimal>] <CRLF>
      REST <SP> <marqueur> <CRLF>
      RNFR <SP> <chemin d'accs> <CRLF>
      RNTO <SP> <chemin d'accs> <CRLF>
      ABOR <CRLF>
      DELE <SP> <chemin d'accs> <CRLF>
      RMD <SP> <chemin d'accs> <CRLF>
      MKD <SP> <chemin d'accs> <CRLF>
      PWD <CRLF>
      LIST [<SP> <chemin d'accs>] <CRLF>
      NLST [<SP> <chemin d'accs>] <CRLF>
      SITE <SP> <chane> <CRLF>
      SYST <CRLF>
      STAT [<SP> <chemin d'accs>] <CRLF>
      HELP [<SP> <chane>] <CRLF>
      NOOP <CRLF>


5.3.2. ARGUMENTS DE COMMANDES FTP

Syntaxe des champs d'arguments ci-avant (notation BNF si applicable) :

 <nom d'utilisateur> ::= <chane>
 <mot de passe> ::= <chane>
 <compte utilisateur> ::= <chane>
 <chane> ::= <char> | <char><chane>
 <char> ::= tout caractre ASCII de code 0  128 sauf <CR> et <LF>
  <marqueur> ::= <pr-chane>
  <pr-chane> ::= <pr-char> | <pr-char><pr-chane>
  <pr-char> ::= caractres imprimables, codes ASCII de 33  126
  <taille d'octet> ::= <nombre>
  <port> ::= <adresse hte>,<numro port>
  <adresse hte> ::= <nombre>,<nombre>,<nombre>,<nombre>
   <numro port> ::= <nombre>,<nombre>
   <nombre> ::= tout entier dcimal entre 1 et 255
    <code format> ::= N | T | C
   <type-code> ::= A [<sp> <code format>] | E [<sp> <code format>] | I | L <sp> <taille d'octet>
   <code structure> ::= F | R | P
   <code mode> ::= S | B | C
   <chemin d'accs> ::= <chane>
    <entier dcimal> ::= tout entier dcimal cod en ASCII
*/

extern int errno;

/**
Structure contenant les information de ftp
Structure containing ftp infos
**/
typedef struct {
char hostname[255]; /**  **/
char ip[13];
unsigned short int port;
char user[25];
char pass[25];
char path[255];
} parsed_url;

int read_answer(int,char*);
int send_cmd(int ,const char *,...);

/**
Convertir le nom de domaine d'un serveur en IP
char *hote=nom de domaine
char *buffer=adresse IP

Converts domain name to IP address
char* hote=domain name
char* buffer=IP address
**/

int getip(char* hote, char*buffer){
	char **addrs;
	struct hostent *infohote;

	infohote=gethostbyname(hote);
	if(!infohote){
		printf("Erreur: %s: unknown host\n",hote);
		return -1;
	}
	if(infohote->h_addrtype!=AF_INET){
		printf("Erreur: %s: n'est pas un hote ip!\n",hote);
		return -1;
	}
	addrs=infohote->h_addr_list;
	//printf("ip: %s\n",inet_ntoa(* (struct in_addr *) *addrs));

	strcpy(buffer,inet_ntoa(* (struct in_addr *) *addrs) );
	return 0;
}

/**
Parse les url du type ftp://user:pass@adresse.ftp.com/repertoire/ et remplit une structure avec les informations
char *url=url ftp
parsed_url *result=structure remplie dans la fonction

Parses ftp urls (ftp://user:pass@ftp.address.com/directory/ and fills a struct with the infos
char *url=url to parse
parsed_url *result=structure filled during the treatment
**/
int parser(char *url, parsed_url *result){
	char *at;
	char *deuxpoints;
	char *tmp;
	char *firstslash;

	/* enleve le 'xxx://' du debut de l'url */
	if(strstr(url,"://")!=NULL){
		tmp=strdup(url);
		sprintf(url,"%s",strstr(url,"://")+3);
		free(tmp);
	}


	/* Choppe un eventuel user et pass */
	/* FIXME! on a vraiment besoin d'une variable temporaire? */
	if( (at=index(url,'@')) ){
		tmp=strdup(url);
		tmp[strlen(tmp)-strlen(at)]='\0';
		/* tmp contient maintenant: 'user:pass' */

		if( (deuxpoints=index(tmp,':'))){
			/* enleve le 'user:pass' du debut de l'url */
			memmove(url,url+strlen(tmp)+1,strlen(url)-strlen(tmp));
			/* rcupre user et pass et remplit la structure */

			strcpy(result->pass,deuxpoints+1);
			*deuxpoints='\0';

			strcpy(result->user,tmp);
		}
	free(tmp);
	}

	else{
		strcpy(result->user,"anonymous");
		strcpy(result->pass,"billg@hotmail.com");
	}

	firstslash=index(url,'/');

	if(firstslash==NULL) firstslash=url+strlen(url);
	tmp=strdup(url);


	tmp[strlen(url)-strlen(firstslash)]='\0';
	if( (deuxpoints=index(tmp,':')) ){
		result->port=atoi(deuxpoints+1);
		*deuxpoints='\0';
		sprintf(url,"%s%s",tmp,firstslash); /* FIXME buffer overflow? */
	}
	else {
		result->port=FTP_DEFAULT_PORT;
	}
	free(tmp);


	/* Si l'URL se termine par un slash, on l'enleve */
	//if(url[strlen(url)-1]=='/') url[strlen(url)-1]='\0';

	/* on separe l'ip et le path */
	tmp=strdup(url);
	firstslash=index(tmp,'/');
	if(firstslash==NULL) firstslash=tmp+strlen(tmp);
	strcpy(result->path,firstslash);
	*firstslash='\0';
	strcpy(result->hostname,tmp);
	free(tmp);

	/* On recupere l'ip */
	getip(result->hostname,result->ip);
	return 1;
}
/**
Connexion au serveur FTP
parsed_url *site=infos pour se connecter au serveur
int *sock=socket avec laquelle on est connect au FTP

Connection to the FTP server
parsed_url *site=information about the server;
int *sock=socket connected to the FTP server
**/
int connection(parsed_url *site, int *sock){
	int retcode;  //return code of funtion calls
	struct sockaddr_in sockaddr;
	int sockaddr_length;
	char *buffer=(char *)malloc(BUFFER_SIZE);

	//socket creation
	*sock=socket(AF_INET,SOCK_STREAM,0);
	sockaddr.sin_family=AF_INET;
	sockaddr.sin_addr.s_addr=inet_addr(site->ip);
	sockaddr.sin_port=htons(site->port);
	sockaddr_length=sizeof(sockaddr);

	//connection to the FTP server
	if( (retcode=connect(*sock, (struct sockaddr*)&sockaddr, sockaddr_length)) ) return retcode;

	//reading server answer
	if((retcode=read_answer(*sock, buffer))!=0) return retcode;

	//FIXME: grer les autres erreurs!!!
	//comparison between 220 and the beginning of the answer (220=Ok, other=problem!)
	//comparaison au code 220 (220=Ok, autre=pb!)
	if((strncmp(buffer,"220",3))!=0) return ERR_FTP_SERVER_CANT_CONNECT;
	printf("\n%s\n\tCONNECTE\n",buffer);
	free(buffer);
	return 0;
}


int read_answer(int sockfd,char *buffer){
   int resultat;
   int i=0;
   char *buffer_ptr=buffer;//=(char *)malloc(sizeof(BUFFER_SIZE));
   //char *tmp;
   fd_set testfds;
   struct timeval timeout;


   timeout.tv_sec=20;
   timeout.tv_usec=0;

   FD_ZERO(&testfds);
   FD_SET(sockfd,&testfds);

 // printf("Dans read_answer avant select \n");
 //  resultat=select(FD_SETSIZE,&testfds, (fd_set*)NULL, (fd_set*)NULL, &timeout);
// printf("Dans read_answer apres select \nresultat=%i\n",resultat);
 /* if(resultat==0){
//buffer_ptr[0]='\0';
	printf("timeout\n");
	return ERR_SOCKET_TIparsed_url *site=infos sur le site FTP ou se connecter
int *sock=socket connectee au serveur FTPMEOUT;
	}
	else if(resultat==-1){
	perror("select");
	//close(sockfd);
	return ERR_SOCKET_SELE	//close(sockfd);CT;
	}*/
	resultat=0;

	//fcntl(sockfd, F_SETFL, O_NONBLOCK);
//	for(;;){

//	while((resultat=read(sockfd,buffer_ptr,BUFFER_SIZE))>0){
	FD_ZERO(&testfds);
   	FD_SET(sockfd,&testfds);

	printf("socket %i\n",sockfd);
	resultat=select(FD_SETSIZE,&testfds, (fd_set*)NULL, (fd_set*)NULL, &timeout);
	printf("resultat du select : %i\n",resultat);
  	if(resultat==0){
		buffer_ptr[0]='\0';
		printf("timeout\n");
	return ERR_SOCKET_TIMEOUT;
	}
	else if(resultat==-1){
		perror("select");
		//close(sockfd);
	return ERR_SOCKET_SELECT;
	}

	resultat=read(sockfd,buffer_ptr,BUFFER_SIZE);
	printf("resultat du read : %i\n",resultat);
	if(resultat<=0) return -1000;   // FIXME: changer la valeur de retour
			buffer_ptr=buffer_ptr+resultat;
			i+=resultat;
		//}


	*buffer_ptr='\0';

	return 0;
	}

/**
Envoie de la commande USER username\r\n au ftp
int sockfd=socket connecte au FTP
char *user=nom d'utilisateur

Sends the USER username\r\n command to the FTP server
int sockfd=socket connected to the FTP
char *user=username
**/
int send_user(int sockfd, char *user){
	int retcode;
	char *buffer;
	if((retcode=send_cmd(sockfd, "USER %s\r\n",user))!=0) return retcode;
	buffer=(char *)malloc(BUFFER_SIZE);
	if((retcode=read_answer(sockfd, buffer))!=0) return retcode;
	printf("Reponse du serveur apres USER:\n%s\n", buffer);
	free(buffer);
	return 0;
}

/**
Envoie de la commande PASS password\r\n au ftp
int sockfd=socket connecte au FTP
char *pass=mot de passe

Sends the PASS password\r\n command to the FTP server
int sockfd=socket connected to the FTP
char *pass=password
**/
int send_pass(int sockfd, char *pass){
	int retcode;
	char *buffer;
	if((retcode=send_cmd(sockfd, "PASS %s\r\n",pass))!=0) return retcode;
	buffer=(char *)malloc(BUFFER_SIZE);
	if((retcode=read_answer(sockfd, buffer))!=0) return retcode;
	printf("Reponse du serveur apres PASS:\n%s\n", buffer);
	free(buffer);
	return 0;

}

/**
Se loguer sur un ftp
parsed_url *site=infos sur le site FTP ou se connecter
int *sock=socket connectee au serveur FTP

Login to the FTP
parsed_url *site=informations about the FTP server to connect to
int *sock=socket connecteed to the FTP server
**/
int login(parsed_url *site, int *sock){
	int retcode;

	if((retcode=connection(site,sock))!=0) return retcode;
		printf("fct connection effectuee\n");
		printf("\t%s est dans user\n",site->user);
		printf("\t%s est dans pass\n",site->pass);
	if((retcode=send_user(*sock,site->user))!=0) return retcode;
	if((retcode=send_pass(*sock,site->pass))!=0) return retcode;
return 0;
}


/**
Fonction d'envoie de commande sur le serveur FTP acceptant des parametres multiples
int sockfd=socket connectee au serveur FTP
const char *fmt,...=chaine de caractere de formatage a la facon printf

Fonction sending FTP commands with multiple args
int sockfd=socket connected to the FTP
const char *fmt,...=format string (printf() style...)
**/
int send_cmd(int sockfd,const char *fmt,...){
	char buffer[8192]; /* FIXME: taille du buffer d'envoi */
	int resultat;
   	fd_set testfds;
   	struct timeval timeout;
	va_list args;
	//int i;


	FD_ZERO(&testfds);
  	FD_SET(sockfd,&testfds);


 	resultat=select(FD_SETSIZE,(fd_set*)NULL, &testfds,  (fd_set*)NULL, &timeout);
	printf("resultat du select dans send_cmd: %i\n",resultat);
  	if(resultat==0){
		printf("timeout\n");
	return ERR_SOCKET_TIMEOUT;
	}
	else if(resultat==-1){
		perror("select");
		//close(sockfd);
	return ERR_SOCKET_SELECT;
	}


   timeout.tv_sec=0;
   timeout.tv_usec=0;

	va_start(args, fmt);
	vsnprintf(buffer,sizeof(buffer)-1,fmt,args);
	va_end(args);

	if( write(sockfd,buffer,strlen(buffer))==-1 ){
		return ERR_SOCKET_WRITE_ERROR;
	}

	return 0;

}
